import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch.utils import data
import VAE
from config import config
import display
from construct import construct
from sklearn.svm import SVC

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

eps = 1e-8

def CEL(x, x_hat):
	sm = nn.Softmax(dim = 1)
	x = sm(x)
	x_hat = sm(x)
	return -(x * torch.log(x_hat + eps) + (1 - x) * torch.log(1 - x_hat + eps)).sum() / x.nelement()

class sampleData(data.Dataset):
	def __init__(self, X, y):
		self.X = X
		self.y = y
		
	def __getitem__(self, index):
		return self.X[index], self.y[index]
	
	def __len__(self):
		return len(self.X)

class Discriminator(nn.Module):
	def __init__(self, in_dim):
		super().__init__()

		self.fc1 = nn.Linear(in_dim, 250)
		self.fc2 = nn.Linear(250, 512)
		self.fc3 = nn.Linear(512, 128)
		self.fc4 = nn.Linear(128, 1)

	def forward(self, x):
		x = torch.sigmoid(self.fc1(x))
		x = torch.sigmoid(self.fc2(x))
		x = torch.sigmoid(self.fc3(x))
		x = self.fc4(x)
		return x

def gradient_penalty(netD, real, fake):
	alpha = torch.rand(real.shape[0], 1).expand(real.shape).to(device)
	interpolates = alpha * real + (1 - alpha) * fake
	interpolates = torch.autograd.Variable(interpolates.to(device), requires_grad=True)
	disc_interpolates = netD(interpolates)
	grads = torch.autograd.grad(inputs = interpolates, outputs = disc_interpolates,
							grad_outputs = torch.ones(disc_interpolates.shape).to(device),
							create_graph = True, retain_graph = True,
							only_inputs = True)[0]
	gp = ((grads.norm(2, dim = 1) - 1) ** 2).mean()
	return gp

def AE(X_train, Y_train, X_test, Y_test, learning_rate, epochs, batch_size):
	dataset = sampleData(X_train, Y_train)
	dataloader = data.DataLoader(dataset, batch_size = batch_size,
		shuffle = True, drop_last = False, **config.kwargs)
	netG = VAE.net(X_train.shape[1], X_train.shape[1]).to(device)
	netD = Discriminator(X_train.shape[1])
	G_optimizer = torch.optim.Adam(netG.parameters(), lr = learning_rate)
	D_optimizer = torch.optim.Adam(netD.parameters(), lr = learning_rate)
	A1 = []
	A2 = []
	for epoch in range(epochs):
		for step, [X, y] in enumerate(dataloader):
			x_hat , mu, log_var, h = netG(X)
			rec_loss = CEL(X, x_hat)
			kl_div = (-0.5 * (log_var + 1 - mu ** 2 - torch.exp(log_var))).sum() / X.shape[0]
			g_loss = rec_loss + kl_div
			G_optimizer.zero_grad()
			g_loss.backward()
			G_optimizer.step()

			fake = torch.Tensor(X.shape).uniform_(-1, 1)
			fake, _, _, _ = netG(fake)
			fake_logits = netD(fake)
			real, _, _, _ = netG(X)
			real_logits = netD(real)
			fake_loss = CEL(fake_logits, torch.zeros_like(fake_logits)).mean()
			real_loss = CEL(real_logits, torch.ones_like(real_logits)).mean()
			gp = gradient_penalty(netD, X, fake)
			d_loss = fake_loss + real_loss + 10. * gp
			D_optimizer.zero_grad()
			d_loss.backward()
			D_optimizer.step()

			# if step == 0:
			# 	print(epoch, 'kl div:%.4f' % kl_div.item(), 'rec loss:%.4f' % rec_loss.item(), 'dloss:%.4f' % d_loss.item())
		torch.save(netG.state_dict(), './backup/netG-01.pt')
		X1, index = construct(X_train, Y_train)
		X2, _ = construct(X_test, Y_test)
		y_output = SVC().fit(X1[:, index], Y_train).predict(X1[:, index])
		TP, TN, FP, FN, P, R, mF1, Acc, G_mean = display.Eval(Y_train, y_output)
		A1.append(Acc)
		y_output = SVC().fit(X1[:, index], Y_train).predict(X2[:, index])
		TP, TN, FP, FN, P, R, mF1, Acc, G_mean = display.Eval(Y_test, y_output)
		A2.append(Acc)

	display.Disp(A1, A2)

	torch.save(netG.state_dict(), './backup/netG-01.pt')
	for step, [X, y] in enumerate(dataloader):
		if (step == 0):
			x_hat, mu, log_var, hid = netG(X)
		else:
			x_hat, mu, log_var, h = netG(X)
			hid = torch.cat([hid, h], axis = 0)
	return hid

def train(X_train, Y_train, X_test, Y_test):
	hid = AE(X_train, Y_train, X_test, Y_test, learning_rate = 0.001, epochs = 50, batch_size = 128)
	# hid = minmaxscaler(hid)
	return hid.detach().cpu().numpy()